#ifndef VERSION_HPP
#define VERSION_HPP

const char* get_osmcoastline_version() noexcept;
const char* get_osmcoastline_long_version() noexcept;
const char* get_libosmium_version() noexcept;

#endif // VERSION_HPP
